@@ -22,7 +22,6 @@ if sys.version_info < (3,):
     # xrange slightly faster than range on python2
     range = xrange
 
-_RE_CLEAN_COMMENT = re.compile(r'(\s*\#.*)', re.VERBOSE)
 _RE_COMMAND = re.compile(r'^\s*(\w+)(\s*)\(', re.VERBOSE)
 _RE_COMMAND_START_SPACES = re.compile(r'^\s*\w+\s*\((\s*)', re.VERBOSE)
 _RE_COMMAND_END_SPACES = re.compile(r'(\s*)\)', re.VERBOSE)
@@ -37,7 +36,9 @@ endmacro
 endwhile
 """.split()
 _USAGE = """
-Syntax: cmakelint.py [--version] [--config=file] [--filter=-x,+y] [--spaces=N] <file> [file] ...
+Syntax: cmakelint.py [--version] [--config=file] [--filter=-x,+y] [--spaces=N]
+                     [--quiet] [--linelength=digits]
+        <file> [file] ...
     filter=-x,+y,...
       Specify a comma separated list of filters to apply
 
@@ -46,9 +47,26 @@ Syntax: cmakelint.py [--version] [--config=file] [--filter=-x,+y] [--spaces=N] <
 
     config=file
       Use the given file for configuration. By default the file
-      $HOME/.cmakelintrc is used if it exists.  Use the value "None" to use no
-      configuration file (./None for a file called literally None)
-      Only the option "filter=" is currently supported in this file.
+      $PWD/.cmakelintrc, ~/.config/cmakelintrc, $XDG_CONFIG_DIR/cmakelintrc or
+      ~/.cmakelintrc is used if it exists. Use the value "None" to use no
+      configuration file (./None for a file called literally None) Only the
+      option "filter=" is currently supported in this file.
+
+    quiet makes output quiet unless errors occurs
+      Mainly used by automation tools when parsing huge amount of files.
+      In those cases actual error might get lost in the pile of other stats
+      prints.
+
+      This argument is also handy for build system integration, so it's
+      possible to add automated lint target to a project and invoke it
+      via build system and have no pollution of terminals or IDE.
+
+    linelength=digits
+      This is the allowed line length for the project. The default value is
+      80 characters.
+
+      Examples:
+        --linelength=120
 
     version
       Show the version number and end
@@ -57,6 +75,7 @@ _ERROR_CATEGORIES = """\
         convention/filename
         linelength
         package/consistency
+        package/stdargs
         readability/logic
         readability/mixedcase
         readability/wonkycase
@@ -68,7 +87,24 @@ _ERROR_CATEGORIES = """\
         whitespace/newline
         whitespace/tabs
 """
-_DEFAULT_CMAKELINTRC = os.path.join(os.environ['HOME'], '.cmakelintrc')
+_DEFAULT_FILENAME = 'CMakeLists.txt'
+
+def DefaultRC():
+    """
+    Check current working directory and XDG_CONFIG_DIR before ~/.cmakelintrc
+    """
+    cwdfile = os.path.join(os.getcwd(), '.cmakelintrc')
+    if os.path.exists(cwdfile):
+        return cwdfile
+    xdg = os.path.join(os.path.expanduser('~'), '.config')
+    if 'XDG_CONFIG_DIR' in os.environ:
+        xdg = os.environ['XDG_CONFIG_DIR']
+    xdgfile = os.path.join(xdg, 'cmakelintrc')
+    if os.path.exists(xdgfile):
+        return xdgfile
+    return os.path.join(os.path.expanduser('~'), '.cmakelintrc')
+
+_DEFAULT_CMAKELINTRC = DefaultRC()
 
 class _CMakeLintState(object):
     def __init__(self):
@@ -76,15 +112,20 @@ class _CMakeLintState(object):
         self.config = 0
         self.errors = 0
         self.spaces = 2
+        self.linelength = 80
         self.allowed_categories = _ERROR_CATEGORIES.split()
+        self.quiet = False
 
     def SetFilters(self, filters):
         if not filters:
             return
-        if self.filters:
-            self.filters.extend(filters.split(','))
+        assert isinstance(self.filters, list)
+        if isinstance(filters, list):
+            self.filters.extend(filters)
+        elif isinstance(filters, str):
+            self.filters.extend([f.strip() for f in filters.split(',') if f])
         else:
-            self.filters = filters.split(',')
+            raise ValueError('Filters should be a list or a comma separated string')
         for f in self.filters:
             if f.startswith('-') or f.startswith('+'):
                 allowed = False
@@ -99,6 +140,12 @@ class _CMakeLintState(object):
     def SetSpaces(self, spaces):
         self.spaces = int(spaces.strip())
 
+    def SetQuiet(self, quiet):
+        self.quiet = quiet
+
+    def SetLineLength(self, linelength):
+        self.linelength = int(linelength)
+
 class _CMakePackageState(object):
     def __init__(self):
         self.sets = []
@@ -110,7 +157,7 @@ class _CMakePackageState(object):
 
     def _GetExpected(self, filename):
         package = os.path.basename(filename)
-        package = re.sub('^Find(.*)\.cmake', lambda m: m.group(1), package)
+        package = re.sub(r'^Find(.*)\.cmake', lambda m: m.group(1), package)
         return package.upper()
 
     def Done(self, filename, errors):
@@ -156,14 +203,46 @@ class _CMakePackageState(object):
 _lint_state = _CMakeLintState()
 _package_state = _CMakePackageState()
 
-def CleanComments(line):
-    return _RE_CLEAN_COMMENT.sub('', line)
+def CleanComments(line, quote=False):
+    """
+    quote means 'was in a quote starting this line' so that
+    quoted lines can be eaten/removed.
+    """
+    if line.find('#') == -1 and line.find('"') == -1:
+        if quote:
+            return '', quote
+        else:
+            return line, quote
+    # else have to check for comment
+    prior = []
+    prev = ''
+    for char in line:
+        try:
+            if char == '"':
+                if prev != '\\':
+                    quote = not quote
+                    prior.append(char)
+                continue
+            elif char == '#' and not quote:
+                break
+            if not quote:
+                prior.append(char)
+        finally:
+            prev = char
+
+    # rstrip removes trailing space between end of command and the comment # start
+
+    return ''.join(prior).rstrip(), quote
 
 class CleansedLines(object):
     def __init__(self, lines):
         self.have_seen_uppercase = None
         self.raw_lines = lines
-        self.lines = [CleanComments(line) for line in lines]
+        self.lines = []
+        quote = False
+        for line in lines:
+            cleaned, quote = CleanComments(line, quote)
+            self.lines.append(cleaned)
 
     def LineNumbers(self):
         return range(0, len(self.lines))
@@ -187,12 +266,13 @@ def CheckLineLength(filename, linenumber, clean_lines, errors):
     Check for lines longer than the recommended length
     """
     line = clean_lines.raw_lines[linenumber]
-    if len(line) > 80:
+    if len(line) > _lint_state.linelength:
         return errors(
                 filename,
                 linenumber,
                 'linelength',
-                'Lines should be <= 80 characters long')
+                'Lines should be <= %d characters long' %
+                    (_lint_state.linelength))
 
 def ContainsCommand(line):
     return _RE_COMMAND.match(line)
@@ -236,6 +316,12 @@ def CheckUpperLowerCase(filename, linenumber, clean_lines, errors):
                         'readability/mixedcase',
                         'Do not mix upper and lower case commands')
 
+def GetInitialSpaces(line):
+    initial_spaces = 0
+    while initial_spaces < len(line) and line[initial_spaces] == ' ':
+        initial_spaces += 1
+    return initial_spaces
+
 def CheckCommandSpaces(filename, linenumber, clean_lines, errors):
     """
     No extra spaces between command and parenthesis
@@ -247,19 +333,26 @@ def CheckCommandSpaces(filename, linenumber, clean_lines, errors):
                 "Extra spaces between '%s' and its ()"%(match.group(1)))
     if match:
         spaces_after_open = len(_RE_COMMAND_START_SPACES.match(line).group(1))
+        initial_spaces = GetInitialSpaces(line)
         initial_linenumber = linenumber
         end = None
         while True:
             line = clean_lines.lines[linenumber]
             end = _RE_COMMAND_END_SPACES.search(line)
+            if end:
+                break
             linenumber += 1
-            if end or linenumber >= len(clean_lines.lines):
+            if linenumber >= len(clean_lines.lines):
                 break
         if linenumber == len(clean_lines.lines) and not end:
             errors(filename, initial_linenumber, 'syntax',
                     'Unable to find the end of this command')
         if end:
             spaces_before_end = len(end.group(1))
+            initial_spaces = GetInitialSpaces(line)
+            if initial_linenumber != linenumber and spaces_before_end >= initial_spaces:
+                spaces_before_end -= initial_spaces
+
             if spaces_after_open != spaces_before_end:
                 errors(filename, initial_linenumber, 'whitespace/mismatch',
                         'Mismatching spaces inside () after command')
@@ -279,10 +372,8 @@ def CheckRepeatLogic(filename, linenumber, clean_lines, errors):
             break
 
 def CheckIndent(filename, linenumber, clean_lines, errors):
-    initial_spaces = 0
     line = clean_lines.raw_lines[linenumber]
-    while initial_spaces < len(line) and line[initial_spaces] == ' ':
-        initial_spaces += 1
+    initial_spaces = GetInitialSpaces(line)
     remainder = initial_spaces % _lint_state.spaces
     if remainder != 0:
         errors(filename, linenumber, 'whitespace/indent',
@@ -307,7 +398,7 @@ def CheckStyle(filename, linenumber, clean_lines, errors):
     CheckRepeatLogic(filename, linenumber, clean_lines, errors)
 
 def CheckFileName(filename, errors):
-    name_match = re.match('Find(.*)\.cmake', os.path.basename(filename))
+    name_match = re.match(r'Find(.*)\.cmake', os.path.basename(filename))
     if name_match:
         package = name_match.group(1)
         if not package.isupper():
@@ -353,6 +444,7 @@ def ProcessLine(filename, linenumber, clean_lines, errors):
       clean_lines CleansedLines instance
       errors      the error handling function
     """
+    CheckLintPragma(filename, linenumber, clean_lines.raw_lines[linenumber], errors)
     CheckLineLength(filename, linenumber, clean_lines, errors)
     CheckUpperLowerCase(filename, linenumber, clean_lines, errors)
     CheckStyle(filename, linenumber, clean_lines, errors)
@@ -363,6 +455,27 @@ def IsValidFile(filename):
     return filename.endswith('.cmake') or os.path.basename(filename).lower() == 'cmakelists.txt'
 
 def ProcessFile(filename):
+    # Store and then restore the filters to prevent pragmas in the file from persisting.
+    original_filters = list(_lint_state.filters)
+    try:
+        return _ProcessFile(filename)
+    finally:
+        _lint_state.filters = original_filters
+
+def CheckLintPragma(filename, linenumber, line, errors=None):
+    # Check this line to see if it is a lint_cmake pragma
+    linter_pragma_start = '# lint_cmake: '
+    if line.startswith(linter_pragma_start):
+        try:
+            _lint_state.SetFilters(line[len(linter_pragma_start):])
+        except ValueError as ex:
+            if errors:
+                errors(filename, linenumber, 'syntax', str(ex))
+        except:
+            print("Exception occurred while processing '{0}:{1}':"
+                  .format(filename, linenumber))
+
+def _ProcessFile(filename):
     lines = ['# Lines start at 1']
     have_cr = False
     if not IsValidFile(filename):
@@ -370,14 +483,16 @@ def ProcessFile(filename):
         return
     global _package_state
     _package_state = _CMakePackageState()
-    CheckFileName(filename, Error)
     for l in open(filename).readlines():
         l = l.rstrip('\n')
         if l.endswith('\r'):
             have_cr = True
             l = l.rstrip('\r')
         lines.append(l)
+        CheckLintPragma(filename, len(lines) - 1, l)
     lines.append('# Lines end here')
+    # Check file name after reading lines incase of a # lint_cmake: pragma
+    CheckFileName(filename, Error)
     if have_cr and os.linesep != '\r\n':
         Error(filename, 0, 'whitespace/newline', 'Unexpected carriage return found; '
                 'better to use only \\n')
@@ -393,9 +508,8 @@ def PrintVersion():
 def PrintUsage(message):
     sys.stderr.write(_USAGE)
     if message:
-        sys.exit('FATAL ERROR: '+message)
-    else:
-        sys.exit(1)
+        sys.stderr.write('FATAL ERROR: %s\n' % message)
+    sys.exit(32)
 
 def PrintCategories():
     sys.stderr.write(_ERROR_CATEGORIES)
@@ -404,6 +518,7 @@ def PrintCategories():
 def ParseOptionFile(contents, ignore_space):
     filters = None
     spaces = None
+    linelength = None
     for line in contents:
         line = line.strip()
         if not line or line.startswith('#'):
@@ -412,14 +527,30 @@ def ParseOptionFile(contents, ignore_space):
             filters = line.replace('filter=', '')
         if line.startswith('spaces='):
             spaces = line.replace('spaces=', '')
+        if line == 'quiet':
+            _lint_state.SetQuiet(True)
+        if line.startswith('linelength='):
+            linelength = line.replace('linelength=', '')
     _lint_state.SetFilters(filters)
     if spaces and not ignore_space:
         _lint_state.SetSpaces(spaces)
+    if linelength is not None:
+        _lint_state.SetLineLength(linelength)
+
+
+# See https://stackoverflow.com/a/30299145 - fixes deprecation warning in py 3.4+
+if sys.version_info[0] == 2:
+    def OpenTextFile(filename):
+        return open(filename, 'rU')
+else:
+    def OpenTextFile(filename):
+        return open(filename, 'r', newline=None)
 
 def ParseArgs(argv):
     try:
         (opts, filenames) = getopt.getopt(argv, '',
-                ['help', 'filter=', 'config=', 'spaces=', 'version'])
+                ['help', 'filter=', 'config=', 'spaces=', 'linelength=',
+                 'quiet', 'version'])
     except getopt.GetoptError:
         PrintUsage('Invalid Arguments')
     filters = ""
@@ -444,9 +575,17 @@ def ParseArgs(argv):
                 ignore_space = True
             except:
                 PrintUsage('spaces expects an integer value')
+        elif opt == '--quiet':
+            _lint_state.quiet = True
+        elif opt == '--linelength':
+            try:
+                _lint_state.SetLineLength(val)
+            except:
+                PrintUsage('line length expects an integer value')
     try:
         if _lint_state.config:
-            try: ParseOptionFile(open(_lint_state.config, 'rU').readlines(), ignore_space)
+            try:
+                ParseOptionFile(OpenTextFile(_lint_state.config).readlines(), ignore_space)
             except IOError:
                 pass
         _lint_state.SetFilters(filters)
@@ -454,7 +593,10 @@ def ParseArgs(argv):
         PrintUsage(str(ex))
 
     if not filenames:
-        PrintUsage('No files were specified!')
+        if os.path.isfile(_DEFAULT_FILENAME):
+            filenames = [_DEFAULT_FILENAME]
+        else:
+            PrintUsage('No files were specified!')
     return filenames
 
 def main():
@@ -462,4 +604,9 @@ def main():
 
     for filename in files:
         ProcessFile(filename)
-    sys.stderr.write("Total Errors: %d\n" % _lint_state.errors)
+    if _lint_state.errors > 0 or not _lint_state.quiet:
+        sys.stderr.write("Total Errors: %d\n" % _lint_state.errors)
+    if _lint_state.errors > 0:
+        return 1
+    else:
+        return 0
